<?php

declare(strict_types=1);

namespace Drupal\entity_field_condition\Plugin\EntityFieldCondition\CompareType;

use Drupal\Core\Annotation\Translation;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\entity_field_condition\Annotation\EntityFieldCompareType;

/**
 * Define the entity field multiple compare type.
 *
 * @EntityFieldCompareType(
 *   id = "entity_field_multiple_compare",
 *   label = @Translation("Multiple")
 * )
 */
class EntityFieldMultipleCompareType extends EntityFieldCompareTypePluginBase {

  /**
   * {@inheritDoc}
   */
  public function defaultConfiguration(): array {
    return [
      'use_strict' => FALSE,
      'allowed_values' => NULL,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritDoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ): array {
    $form['allowed_values'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Allowed Values'),
      '#required' => TRUE,
      '#description' => $this->t(
        'Input a comma separated list of allowed values.'
      ),
      '#default_value' => $this->getRawAllowedValues(),
    ];
    $form['use_strict'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Use Strict'),
      '#description' => $this->t(
        'If checked, the value will be evaluated using strict mode.'
      ),
      '#default_value' => $this->getUseStrict(),
    ];

    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function evaluate(
    ContentEntityInterface $entity,
    string $field_name
  ): bool {
    return in_array(
      $entity->get($field_name)->getString(),
      $this->getAllowedValues(),
      $this->getUseStrict()
    );
  }

  /**
   * Get allowed values.
   *
   * @return array
   *   An array of allowed values.
   */
  protected function getAllowedValues(): array {
    return array_map(
      'trim',
      explode(',', $this->getRawAllowedValues())
    );
  }

  /**
   * Get the use strict mode.
   *
   * @return bool
   *   The use strict mode.
   */
  protected function getUseStrict(): bool {
    return (bool) ($this->getConfiguration()['use_strict'] ?? FALSE);
  }

  /**
   * Get the raw allowed values.
   *
   * @return string|null
   *   The raw allowed values string.
   */
  protected function getRawAllowedValues(): ?string {
    return $this->getConfiguration()['allowed_values'] ?? NULL;
  }

}
